@@ -23,17 +23,19 @@
 -- > main = do
 -- >      request <- parseUrl "http://google.com/"
 -- >      withManager $ \manager -> do
--- >          Response _ _ src <- http request manager
--- >          src C.$$ sinkFile "google.html"
+-- >          Response _ _ _ src <- http request manager
+-- >          src C.$$+- sinkFile "google.html"
 --
 -- The following headers are automatically set by this module, and should not
 -- be added to 'requestHeaders':
 --
 -- * Content-Length
 --
--- * Host
---
--- * Accept-Encoding (not currently set, but client usage of this variable /will/ cause breakage).
+-- Note: In previous versions, the Host header would be set by this module in
+-- all cases. Starting from 1.6.1, if a Host header is present in
+-- @requestHeaders@, it will be used in place of the header this module would
+-- have generated. This can be useful for calling a server which utilizes
+-- virtual hosting.
 --
 -- Any network code on Windows requires some initialization, and the network
 -- library provides withSocketsDo to perform it. Therefore, proper usage of
@@ -49,6 +51,22 @@
 -- >      $ simpleHttp "http://www.haskell.org/" >>= L.putStr
 -- >
 -- > Cookies are implemented according to RFC 6265.
+--
+-- Note that by default, the functions in this package will throw exceptions
+-- for non-2xx status codes. If you would like to avoid this, you should use
+-- 'checkStatus', e.g.:
+--
+-- > import Data.Conduit.Binary (sinkFile)
+-- > import Network.HTTP.Conduit
+-- > import qualified Data.Conduit as C
+-- > import Network
+-- >
+-- > main :: IO ()
+-- > main = withSocketsDo $ do
+-- >      request' <- parseUrl "http://www.yesodweb.com/does-not-exist"
+-- >      let request = request' { checkStatus = \_ _ -> Nothing }
+-- >      res <- withManager $ httpLbs request
+-- >      print res
 module Network.HTTP.Conduit
     ( -- * Perform a request
       simpleHttp
@@ -63,6 +81,7 @@ module Network.HTTP.Conduit
     , def
     , method
     , secure
+    , clientCertificates
     , host
     , port
     , path
@@ -75,15 +94,18 @@ module Network.HTTP.Conduit
     , decompress
     , redirectCount
     , checkStatus
+    , responseTimeout
       -- * Manager
     , Manager
     , newManager
     , closeManager
     , withManager
+    , withManagerSettings
       -- ** Settings
     , ManagerSettings
     , managerConnCount
     , managerCheckCerts
+    , managerCertStore
       -- *** Defaults
     , defaultCheckCerts
       -- * Cookies
@@ -108,6 +130,8 @@ module Network.HTTP.Conduit
     , alwaysDecompress
     , browserDecompress
       -- * Request bodies
+      -- | "Network.HTTP.Conduit.MultipartFormData" provides an API for building
+      -- form-data request bodies.
     , urlEncodedBody
       -- * Exceptions
     , HttpException (..)
@@ -124,14 +148,18 @@ import qualified Network.HTTP.Types as W
 import Data.Default (def)
 
 import Control.Exception.Lifted (throwIO)
+import Control.Monad ((<=<))
 import Control.Monad.IO.Class (MonadIO (liftIO))
-import Control.Monad.Trans.Control (MonadBaseControl)
+import Control.Monad.Trans.Resource
+import Control.Monad.Trans.State (get, put, evalStateT)
+import Control.Monad.Trans (lift)
 
+import Control.Exception (fromException, toException)
 import qualified Data.Conduit as C
+import qualified Data.Conduit.Binary as CB
 import qualified Data.Conduit.Internal as CI
 import Data.Conduit.Blaze (builderToByteString)
-import Data.Conduit (MonadResource)
-import Control.Exception.Lifted (try, SomeException)
+import Control.Exception.Lifted (try, IOException)
 
 import Data.Time.Clock
 
@@ -140,6 +168,7 @@ import Network.HTTP.Conduit.Response
 import Network.HTTP.Conduit.Manager
 import Network.HTTP.Conduit.ConnInfo
 import Network.HTTP.Conduit.Cookies
+import Network.HTTP.Conduit.Internal (httpRedirect)
 
 -- | The most low-level function for initiating an HTTP request.
 --
@@ -164,7 +193,7 @@ http
     :: (MonadResource m, MonadBaseControl IO m)
     => Request m
     -> Manager
-    -> m (Response (C.Source m S.ByteString))
+    -> m (Response (C.ResumableSource m S.ByteString))
 http req0 manager = do
     res@(Response status _version hs body) <-
         if redirectCount req0 == 0
@@ -173,27 +202,45 @@ http req0 manager = do
     case checkStatus req0 status hs of
         Nothing -> return res
         Just exc -> do
-            CI.runFinalize $ CI.pipeClose body
-            liftIO $ throwIO exc
+            exc' <-
+                case fromException exc of
+                    Just (StatusCodeException s hdrs) -> do
+                        lbs <- body C.$$+- CB.take 1024
+                        return $ toException $ StatusCodeException s $ hdrs ++
+                            [("X-Response-Body-Start", S.concat $ L.toChunks lbs)]
+                    _ -> do
+                        let CI.ResumableSource _ final = body
+                        final
+                        return exc
+            liftIO $ throwIO exc'
   where
-    go 0 _ _ = liftIO $ throwIO TooManyRedirects
-    go count req'' cookie_jar'' = do
+    go count req''' cookie_jar''' = (`evalStateT` cookie_jar''') $
+      httpRedirect
+      count
+      (\req'' -> do
+        cookie_jar'' <- get
         now <- liftIO getCurrentTime
         let (req', cookie_jar') = insertCookiesIntoRequest req'' (evictExpiredCookies cookie_jar'' now) now
-        res <- httpRaw req' manager
+        res <- lift $ httpRaw req' manager
         let (cookie_jar, _) = updateCookieJar res req' now cookie_jar'
-        case getRedirectedRequest req' (responseHeaders res) (W.statusCode (responseStatus res)) of
-            Just req -> go (count - 1) req cookie_jar
-            Nothing -> return res
+        put cookie_jar
+        let mreq = getRedirectedRequest req' (responseHeaders res) (W.statusCode (responseStatus res))
+        return (res, mreq))
+      lift
+      req'''
 
 -- | Get a 'Response' without any redirect following.
 httpRaw
      :: (MonadBaseControl IO m, MonadResource m)
      => Request m
      -> Manager
-     -> m (Response (C.Source m S.ByteString))
+     -> m (Response (C.ResumableSource m S.ByteString))
 httpRaw req m = do
-    (connRelease, ci, isManaged) <- getConn req m
+    (connRelease, ci, isManaged) <- getConnectionWrapper
+        req
+        (responseTimeout req)
+        (failedConnectionException req)
+        (getConn req m)
     let src = connSource ci
 
     -- Originally, we would only test for exceptions when sending the request,
@@ -215,7 +262,7 @@ httpRaw req m = do
         -- thrown in the response body, just throw them as normal.
         (Right x, _) -> return x
   where
-    try' :: MonadBaseControl IO m => m a -> m (Either SomeException a)
+    try' :: MonadBaseControl IO m => m a -> m (Either IOException a)
     try' = try
 
 -- | Download the specified 'Request', returning the results as a 'Response'.
@@ -225,7 +272,7 @@ httpRaw req m = do
 -- interleaved actions on the response body during download, you'll need to use
 -- 'http' directly. This function is defined as:
 --
--- @httpLbs = 'lbsResponse' . 'http'@
+-- @httpLbs = 'lbsResponse' <=< 'http'@
 --
 -- Even though the 'Response' contains a lazy bytestring, this
 -- function does /not/ utilize lazy I/O, and therefore the entire
@@ -236,7 +283,7 @@ httpRaw req m = do
 -- Note: Unlike previous versions, this function will perform redirects, as
 -- specified by the 'redirectCount' setting.
 httpLbs :: (MonadBaseControl IO m, MonadResource m) => Request m -> Manager -> m (Response L.ByteString)
-httpLbs r = lbsResponse . http r
+httpLbs r = lbsResponse <=< http r
 
 -- | Download the specified URL, following any redirects, and
 -- return the response body.
@@ -244,13 +291,15 @@ httpLbs r = lbsResponse . http r
 -- This function will 'throwIO' an 'HttpException' for any
 -- response with a non-2xx status code (besides 3xx redirects up
 -- to a limit of 10 redirects). It uses 'parseUrl' to parse the
--- input. This function essentially wraps 'httpLbsRedirect'.
+-- input. This function essentially wraps 'httpLbs'.
 --
 -- Note: Even though this function returns a lazy bytestring, it
 -- does /not/ utilize lazy I/O, and therefore the entire response
 -- body will live in memory. If you want constant memory usage,
--- you'll need to use the @conduit@ package and 'http' or
--- 'httpRedirect' directly.
+-- you'll need to use the @conduit@ package and 'http' directly.
+--
+-- Note: This function creates a new 'Manager'. It should be avoided
+-- in production code.
 simpleHttp :: MonadIO m => String -> m L.ByteString
 simpleHttp url = liftIO $ withManager $ \man -> do
     url' <- liftIO $ parseUrl url
