@@ -12,12 +12,16 @@ module Network.HTTP.Conduit.Manager
     , getConn
     , ConnReuse (..)
     , withManager
+    , withManagerSettings
     , ConnRelease
     , ManagedConn (..)
     , defaultCheckCerts
+    , failedConnectionException
     ) where
 
+#if !MIN_VERSION_base(4,6,0)
 import Prelude hiding (catch)
+#endif
 import Data.Monoid (mappend)
 import System.IO (hClose, hFlush, IOMode(..))
 import qualified Data.IORef as I
@@ -35,8 +39,9 @@ import Control.Monad.Trans.Control (MonadBaseControl)
 import Control.Monad.IO.Class (MonadIO, liftIO)
 import Control.Exception (mask_, SomeException, catch)
 import Control.Monad.Trans.Resource
-    ( ResourceT, runResourceT, MonadResource (..)
+    ( ResourceT, runResourceT, MonadResource
     , MonadThrow, MonadUnsafeIO
+    , allocate, resourceMask, register, release
     )
 import Control.Concurrent (forkIO, threadDelay)
 import Data.Time (UTCTime (..), Day (..), DiffTime, getCurrentTime, addUTCTime)
@@ -45,14 +50,16 @@ import Control.DeepSeq (deepseq)
 import Network (connectTo, PortID (PortNumber), HostName)
 import Network.Socket (socketToHandle)
 import Data.Certificate.X509 (X509, encodeCertificate)
+import Data.CertificateStore (CertificateStore)
+import System.Certificate.X509 (getSystemCertificateStore)
 
-import qualified Network.HTTP.Types as W
+import Network.TLS (PrivateKey)
 import Network.TLS.Extra (certificateVerifyChain, certificateVerifyDomain)
 
 import Network.HTTP.Conduit.ConnInfo
+import Network.HTTP.Conduit.Types
 import Network.HTTP.Conduit.Util (hGetSome)
 import Network.HTTP.Conduit.Parser (parserHeadersFromByteString)
-import Network.HTTP.Conduit.Request
 import Network.Socks5 (SocksConf, socksConnectWith)
 import Data.Default
 import Data.Maybe (mapMaybe)
@@ -63,8 +70,10 @@ import System.IO (Handle)
 data ManagerSettings = ManagerSettings
     { managerConnCount :: Int
       -- ^ Number of connections to a single host to keep alive. Default: 10.
-    , managerCheckCerts :: W.Ascii -> [X509] -> IO TLSCertificateUsage
+    , managerCheckCerts :: CertificateStore -> S8.ByteString -> [X509] -> IO CertificateUsage
       -- ^ Check if the server certificate is valid. Only relevant for HTTPS.
+    , managerCertStore :: IO CertificateStore
+      -- ^ Load up the certificate store. By default uses the system store.
     }
 
 type X509Encoded = L.ByteString
@@ -73,25 +82,26 @@ instance Default ManagerSettings where
     def = ManagerSettings
         { managerConnCount = 10
         , managerCheckCerts = defaultCheckCerts
+        , managerCertStore = getSystemCertificateStore
         }
 
 -- | Check certificates using the operating system's certificate checker.
-defaultCheckCerts :: W.Ascii -> [X509] -> IO TLSCertificateUsage
-defaultCheckCerts host' certs =
+defaultCheckCerts :: CertificateStore -> S8.ByteString -> [X509] -> IO CertificateUsage
+defaultCheckCerts certStore host' certs =
     case certificateVerifyDomain (S8.unpack host') certs of
-        CertificateUsageAccept -> certificateVerifyChain certs
+        CertificateUsageAccept -> certificateVerifyChain certStore certs
         rejected               -> return rejected
 
--- | Keeps track of open connections for keep-alive.  May be used
--- concurrently by multiple threads.
+-- | Keeps track of open connections for keep-alive.
+-- If possible, you should share a single 'Manager' between multiple threads and requests.
 data Manager = Manager
     { mConns :: !(I.IORef (Maybe (Map.Map ConnKey (NonEmptyList ConnInfo))))
     -- ^ @Nothing@ indicates that the manager is closed.
     , mMaxConns :: !Int
     -- ^ This is a per-@ConnKey@ value.
-    , mCheckCerts :: W.Ascii -> [X509] -> IO TLSCertificateUsage
+    , mCheckCerts :: S8.ByteString -> [X509] -> IO CertificateUsage
     -- ^ Check if a certificate is valid.
-    , mCertCache :: !(I.IORef (Map.Map W.Ascii (Map.Map X509Encoded UTCTime)))
+    , mCertCache :: !(I.IORef (Map.Map S8.ByteString (Map.Map X509Encoded UTCTime)))
     -- ^ Cache of validated certificates. The @UTCTime@ gives the expiration
     -- time for the validity of the certificate. The @Ascii@ is the hostname.
     }
@@ -140,16 +150,28 @@ addToList now maxCount x l@(Cons _ currCount _ _)
     | otherwise = (l, Just x)
 
 -- | Create a 'Manager'. You must manually call 'closeManager' to shut it down.
+--
+-- Creating a new 'Manager' is an expensive operation, you are advised to share
+-- a single 'Manager' between requests instead.
 newManager :: ManagerSettings -> IO Manager
 newManager ms = do
+    icertStore <- I.newIORef Nothing
+    let getCertStore = do
+            mcertStore <- I.readIORef icertStore
+            case mcertStore of
+                Nothing -> do
+                    certStore <- managerCertStore ms
+                    I.writeIORef icertStore $ Just certStore
+                    return certStore
+                Just x -> return x
     mapRef <- I.newIORef (Just Map.empty)
     certCache <- I.newIORef Map.empty
     _ <- forkIO $ reap mapRef certCache
-    return $ Manager mapRef (managerConnCount ms) (managerCheckCerts ms) certCache
+    return $ Manager mapRef (managerConnCount ms) (\x y -> getCertStore >>= \cs -> managerCheckCerts ms cs x y) certCache
 
 -- | Collect and destroy any stale connections.
 reap :: I.IORef (Maybe (Map.Map ConnKey (NonEmptyList ConnInfo)))
-     -> I.IORef (Map.Map W.Ascii (Map.Map X509Encoded UTCTime))
+     -> I.IORef (Map.Map S8.ByteString (Map.Map X509Encoded UTCTime))
      -> IO ()
 reap mapRef certCacheRef =
     mask_ loop
@@ -237,7 +259,7 @@ neFromList xs =
 -- | Create a new manager, use it in the provided function, and then release it.
 --
 -- This function uses the default manager settings. For more control, use
--- 'newManager'.
+-- 'withManagerSettings'.
 withManager :: ( MonadIO m
                , MonadBaseControl IO m
                , MonadThrow m
@@ -247,6 +269,16 @@ withManager f = runResourceT $ do
     (_, manager) <- allocate (newManager def) closeManager
     f manager
 
+-- | Create a new manager with provided settings, use it in the provided function, and then release it.
+withManagerSettings :: ( MonadIO m
+                       , MonadBaseControl IO m
+                       , MonadThrow m
+                       , MonadUnsafeIO m
+                       ) => ManagerSettings -> (Manager -> ResourceT m a) -> m a
+withManagerSettings s f = runResourceT $ do
+    (_, manager) <- allocate (newManager s) closeManager
+    f manager
+
 -- | Close all connections in a 'Manager'. Afterwards, the
 -- 'Manager' can be reused if desired.
 closeManager :: Manager -> IO ()
@@ -278,21 +310,23 @@ socketDesc :: String -> Int -> String -> String
 socketDesc h p t = unwords [h, show p, t]
 
 getSslConn :: MonadResource m
-            => ([X509] -> IO TLSCertificateUsage)
+            => ([X509] -> IO CertificateUsage)
+            -> [(X509, Maybe PrivateKey)]
             -> Manager
             -> String -- ^ host
             -> Int -- ^ port
             -> Maybe SocksConf -- ^ optional socks proxy
             -> m (ConnRelease m, ConnInfo, ManagedConn)
-getSslConn checkCert man host' port' socksProxy' =
+getSslConn checkCert clientCerts man host' port' socksProxy' =
     getManagedConn man (ConnKey (T.pack host') port' True) $
-        (connectionTo host' (PortNumber $ fromIntegral port') socksProxy' >>= sslClientConn desc checkCert)
+        (connectionTo host' (PortNumber $ fromIntegral port') socksProxy' >>= sslClientConn desc host' checkCert clientCerts)
   where
     desc = socketDesc host' port' "secured"
 
 getSslProxyConn
             :: MonadResource m
-            => ([X509] -> IO TLSCertificateUsage)
+            => ([X509] -> IO CertificateUsage)
+            -> [(X509, Maybe PrivateKey)]
             -> S8.ByteString -- ^ Target host
             -> Int -- ^ Target port
             -> Manager
@@ -300,9 +334,9 @@ getSslProxyConn
             -> Int -- ^ Proxy port
             -> Maybe SocksConf -- ^ optional SOCKS proxy
             -> m (ConnRelease m, ConnInfo, ManagedConn)
-getSslProxyConn checkCert thost tport man phost pport socksProxy' =
+getSslProxyConn checkCert clientCerts thost tport man phost pport socksProxy' =
     getManagedConn man (ConnKey (T.pack phost) pport True) $
-        doConnect >>= sslClientConn desc checkCert
+        doConnect >>= sslClientConn desc phost checkCert clientCerts
   where
     desc = socketDesc phost pport "secured-proxy"
     doConnect = do
@@ -325,8 +359,6 @@ getSslProxyConn checkCert thost tport man phost pport socksProxy' =
         error $ "Proxy failed to CONNECT to '"
                 ++ S8.unpack thost ++ ":" ++ show tport ++ "' : " ++ s
 
-data ManagedConn = Fresh | Reused
-
 -- | This function needs to acquire a @ConnInfo@- either from the @Manager@ or
 -- via I\/O, and register it with the @ResourceT@ so it is guaranteed to be
 -- either released or returned to the manager.
@@ -372,9 +404,19 @@ getManagedConn man key open = resourceMask $ \restore -> do
             release releaseKey
     return (connRelease, ci, isManaged)
 
-data ConnReuse = Reuse | DontReuse
+-- | Create an exception to be thrown if the connection for the given request
+-- fails.
+failedConnectionException :: Request m -> HttpException
+failedConnectionException req =
+    FailedConnectionException host' port'
+  where
+    (_, host', port') = getConnDest req
 
-type ConnRelease m = ConnReuse -> m ()
+getConnDest :: Request m -> (Bool, String, Int)
+getConnDest req =
+    case proxy req of
+        Just p -> (True, S8.unpack (proxyHost p), proxyPort p)
+        Nothing -> (False, S8.unpack $ host req, port req)
 
 getConn :: MonadResource m
         => Request m
@@ -384,17 +426,14 @@ getConn req m =
     go m connhost connport (socksProxy req)
   where
     h = host req
-    (useProxy, connhost, connport) =
-        case proxy req of
-            Just p -> (True, S8.unpack (proxyHost p), proxyPort p)
-            Nothing -> (False, S8.unpack h, port req)
+    (useProxy, connhost, connport) = getConnDest req
     go =
         case (secure req, useProxy) of
             (False, _) -> getSocketConn
-            (True, False) -> getSslConn $ checkCerts m h
-            (True, True) -> getSslProxyConn (checkCerts m h) h (port req)
+            (True, False) -> getSslConn (checkCerts m h) (clientCertificates req)
+            (True, True) -> getSslProxyConn (checkCerts m h) (clientCertificates req) h (port req)
 
-checkCerts :: Manager -> W.Ascii -> [X509] -> IO TLSCertificateUsage
+checkCerts :: Manager -> S8.ByteString -> [X509] -> IO CertificateUsage
 checkCerts man host' certs = do
 #if DEBUG
     putStrLn $ "checkCerts for host: " ++ show host'
